package de.tu_darmstadt.tk.SmartHomeNetworkSim.core;

import java.io.BufferedWriter;
import java.io.File;
import java.io.FileWriter;
import java.io.IOException;
import java.util.Collection;
import java.util.LinkedList;
import java.util.Observable;

import javax.swing.Timer;

import de.tu_darmstadt.tk.SmartHomeNetworkSim.control.Controller;
import de.tu_darmstadt.tk.SmartHomeNetworkSim.core.util.PacketComparator;

/**
 * Manages the Simulation by calling the relevant methods.
 *
 * @author Andreas T. Meyer-Berg
 */
public class SimulationManager extends Observable {
	/**
	 * Model which should be simulated
	 */
	private Model model;

	/**
	 * Controller for manipulationAlgorithms
	 */
	private Controller controller;

	/**
	 * True if packets should be printed
	 */
	private boolean printPackets = false;

	/**
	 * Whether the model status changed during the last simulation and the panel
	 * should be repainted
	 */
	private boolean statusChanged = false;

	/**
	 * Writer to write the packets to a file
	 */
	private BufferedWriter writer;

	/**
	 * Base File for the exported Packages
	 */
	private File exportFile = new File("testPackets.log");

	/**
	 * Split Links into differentFiles
	 */
	private boolean splitLinkExportFiles = false;

	/**
	 * Timer which triggers each simulation step
	 */
	private Timer timer;

	/**
	 * First Timestep of the simulation
	 */
	private long startTime = 0L;
	
	/**
	 * Last Timestep of the simulation
	 */
	private long endTime = 1000L;
	
	/**
	 * Current TimeStep in the simulation
	 */
	private long currentTime = 0L;
	
	/**
	 * Duration of the simulation
	 */
	private long duration = 100L;
	
	/**
	 * ManipulationAlgorithm Instances, which should run in each timestep
	 */
	private LinkedList<NetworkManipulationAlgorithm> algos = new LinkedList<NetworkManipulationAlgorithm>();

	/**
	 * Creates a new Simulationmanager
	 * 
	 * @param model
	 *            Model that should be simulated
	 */
	public SimulationManager(Model model) {
		this.model = model;
		this.controller = null;
		timer = new Timer(0, t -> simulateTimeStep());
	}

	/**
	 * Simulate the next Timestep
	 */
	private void simulateTimeStep() {
		// Just simulate if timer is running
		if (endTime <= currentTime)
			stopSimulation();
		if (currentTime + duration <= endTime) {
			simulateTimeIntervall(currentTime, duration);
			currentTime += duration;
			notifyPanels();
		} else {
			simulateTimeIntervall(currentTime, endTime - currentTime);
			currentTime = endTime;
			stopSimulation();
		}
	}

	/**
	 * Simulates one time step at a given time fur a given duration
	 * 
	 * @param startTime
	 *            Time the simulation interval starts in
	 *            System.currentTimeMillis() time
	 * @param duration
	 *            Duration of the simulation interval in milliseconds
	 */
	public void simulateTimeIntervall(long startTime, long duration) {
		// Nothing changed so far
		statusChanged = false;
		// Simulate all Links, and their connections
		model.getConnectionNetworks().forEach(d -> {
			d.simulateTimeInterval(startTime, duration);
			if (d.getStatusChanged()) {
				for (Connection c : d.getConnections()) {
					if (c.getStatus() == Connection.DONE) {
						model.getConnections().remove(c);
					}
				}
				statusChanged = true;
			}
		});
		// Simulate SmartDevices - if they need some logic
		model.getDevices().forEach(d -> d.simulateTimeStep(startTime, duration));
		// Store Packages/Export Packages etc. (for debug purposes)
		if (statusChanged) {
			model.setChanged();
			model.notifyObservers();
		}
		runAlgorithms(startTime);
		exportPacketsOfLastTimeStep();
	}

	/**
	 * Exports the packets which were sent during the last time step according
	 * to the settings (whether packages should be printed, specified export
	 * File, splitLinks setting etc.)
	 */
	public void exportPacketsOfLastTimeStep() {
		if (printPackets) {
			if (!splitLinkExportFiles) {
				/**
				 * Packets of all links, merged together
				 */
				LinkedList<Packet> packets = new LinkedList<Packet>();
				model.getConnectionNetworks().forEach(a->packets.addAll(a.getPackets()));
				packets.sort(new PacketComparator());
				
				writePacketsToFile(packets, exportFile);
			}else{
				for(Link link:model.getConnectionNetworks()){
					/**
					 * File name:
					 * test.log
					 * to
					 * test_linkName.log
					 */
					String fileName = exportFile.getName();
					//Add linkName
					if(fileName.contains(".")){
						String baseName = fileName.substring(0, fileName.lastIndexOf('.'));
						String ending = fileName.substring(fileName.lastIndexOf('.')+1, fileName.length());
						fileName = baseName +"_"+link.getName()+"."+ending;
					}else{
						fileName = fileName +"_"+link.getName();
					}
					/**
					 * Path to parent directory
					 */
					String parent = exportFile.getParent();
					/**
					 * File with the fileName
					 */
					File linkExportFile = new File((parent == null?"":parent)+fileName);
					/**
					 * Write Packets
					 */
					writePacketsToFile(link.getPackets(), linkExportFile);
				}
			}
		}
	}

	/**
	 * Writes all packets of the given Collection to the specified file
	 * @param packets packets which should be written
	 * @param file file the packets should be written to
	 */
	private void writePacketsToFile(Collection<Packet> packets, File file){
		if(file == null || packets == null)
			return;
		try {
			//Check if file exists, if not create
			file.createNewFile();
			//Start Writing
			writer = new BufferedWriter(new FileWriter(file.getAbsolutePath(), true));
			//Write all packets
			packets.forEach(p -> {
				try {
					writer.append(p.getTextualRepresentation() + "\n");
				} catch (Exception e) {
					System.err.println("Warning: Error on exporting packets: "+e.toString());
				}
			});
		} catch (Exception e) {
			System.err.println("Warning: Writing failed on file "+file.getPath()+"\n"+e.toString());
		} finally {
			//Close open writer
			if (writer != null)
				try {
					writer.close();
				} catch (IOException e) {}
		}
	}

	/**
	 * Returns true if the simulations will print the packets
	 * 
	 * @return if the packets are printed
	 */
	public boolean getPrintPackets() {
		return printPackets;
	}

	/**
	 * Sets Print Packets, if true, the simulation will print the packets
	 * 
	 * @param printPackets
	 *            true if simulation should print the packets
	 */
	public void setPrintPackets(boolean printPackets) {
		this.printPackets = printPackets;
		notifyPanels();
	}

	/**
	 * Get BaseFile of packet exports
	 * 
	 * @return ExportFile File, where the packets are written to
	 */
	public File getExportFile() {
		return exportFile;
	}

	/**
	 * Set BaseFile for the packet export
	 * 
	 * @param exportFile
	 *            File, where the packets are written to
	 */
	public void setExportFile(File exportFile) {
		this.exportFile = exportFile;
		notifyPanels();
	}

	/**
	 * True if each links uses a separated File: e.g. exportFile_link.log
	 * 
	 * @return true, if each links uses a single file for export
	 */
	public boolean isSplitLinkExportFiles() {
		return splitLinkExportFiles;
	}

	/**
	 * Set whether packets should be split into different files for each link
	 * 
	 * @param splitLinkExportFiles
	 *            true if each link should export to a single file
	 */
	public void setSplitLinkExportFiles(boolean splitLinkExportFiles) {
		this.splitLinkExportFiles = splitLinkExportFiles;
		notifyPanels();
	}

	/**
	 * Start the simulation
	 */
	public void startSimulation() {
		timer.start();
		notifyPanels();
	}

	/**
	 * Stop the simulation
	 */
	public void stopSimulation() {
		timer.stop();
		notifyPanels();
	}

	/**
	 * Resets the Simulation to the start time
	 */
	public void resetSimulation() {
		timer.stop();
		timer = new Timer(0, a -> simulateTimeStep());
		currentTime = startTime;
		resetSimulation(currentTime);
		notifyPanels();
	}

	/**
	 * Reset Simulation
	 * 
	 * @param timestep new timestep of ports after the reset
	 */
	public void resetSimulation(long timestep) {
		for (SmartDevice d : model.getDevices())
			for (Port p : d.getPorts())
				if (p.getLastTrigger() > timestep)
					p.setLastTrigger(timestep);
	}

	/**
	 * Returns true if the simulation is running, false if not
	 * 
	 * @return true if running
	 */
	public boolean isRunning() {
		return timer.isRunning();
	}

	/**
	 * Returns the StartTime of the simulation
	 * 
	 * @return startTime
	 */
	public long getStartTime() {
		return startTime;
	}

	/**
	 * Sets the new startTime
	 * 
	 * @param startTime
	 *            time the simulations starts
	 */
	public void setStartTime(long startTime) {
		this.startTime = startTime;
		notifyPanels();
	}

	/**
	 * Returns the end time of the simulation
	 * 
	 * @return End time of the simulation
	 */
	public long getEndTime() {
		return endTime;
	}

	/**
	 * Sets the new startTime
	 * @param endTime new EndTime
	 */
	public void setEndTime(long endTime) {
		this.endTime = endTime;
		notifyPanels();
	}

	/**
	 * @return the currentTime
	 */
	public long getCurrentTime() {
		return currentTime;
	}

	/**
	 * @param currentTime
	 *            the currentTime to set
	 */
	public void setCurrentTime(long currentTime) {
		this.currentTime = currentTime;
		notifyPanels();
	}

	/**
	 * Returns the simulation step duration in milliseconds
	 * 
	 * @return duration of each simulation step in milliseconds
	 */
	public long getStepDuration() {
		return duration;
	}

	/**
	 * Sets the duration of simulation steps
	 * 
	 * @param duration
	 *            duration in milliseconds of a step
	 */
	public void setStepDuration(long duration) {
		this.duration = duration;
		notifyPanels();
	}

	/**
	 * Notify the panels, which could update their GUI
	 */
	private void notifyPanels() {
		this.setChanged();
		this.notifyObservers();
	}
	
	/**
	 * Runs all registered algorithms at the currentTimeStep
	 * @param time currentTime of the simulation
	 */
	public void runAlgorithms(long time) {
		/**
		 * Run all Algorithms
		 */
		for(NetworkManipulationAlgorithm algo:algos){
			algo.runAlgorithm(controller, time);
		}
	}
	/**
	 * Returns all registered algorithms of the simulation
	 * @return all registered algorithms
	 */
	public LinkedList<NetworkManipulationAlgorithm> getAlgorithms(){
		return algos;
	}
	/**
	 * Adds an algorithm, which should be executed each timestep
	 * @param algo new algorithm
	 */
	public void addAlgorithm(NetworkManipulationAlgorithm algo, Controller controller){
		this.controller = controller;
		if(algo!=null)
			algos.add(algo);
	}
	
	/**
	 * Removes algorithms from the simulation
	 * @param algo algorithm to be removed
	 */
	public void removeAlgo(NetworkManipulationAlgorithm algo){
		algos.remove(algo);
	}
}
